import rethinkdb as r
from rethinkdb.ast import RqlQuery


class TableNotFound(Exception):
    """Indicates that a table wasn't found """
    pass


class RethinkDBDriver(object):
    def __init__(self, connection_settings):
        """RethinkDB Driver to wrap connection management on top of the
        official driver.

        @param connection_settings: Connection settings to be used to create
        connection.
        """
        if not isinstance(connection_settings, dict):
            raise ValueError('The argument connection_settings should be a '
                             '<dict>. Got <%s> instead'
                             % type(connection_settings))

        self._conn = None
        self.connection_settings = connection_settings

    @property
    def connection(self):
        """Gets/creates connection.

        @returns Connection
        """
        if not self._conn:
            self._conn = r.connect(**self.connection_settings)

        return self._conn

    def table_exists(self, table_name):
        """Checks if a table exists in the database

        @param table_name: table name
        @returns Whether the table was found in the db or not.
        """
        table_list = self.execute(r.table_list())
        return table_name in table_list

    def get_table(self, table_name):
        """Gets table by name.

        @param table_name: table name to be retrieved from the current
        connection.
        @returns Table with the requested name.
        @raises TableNotFound: The table wasn't found in the current
        connection.
        """
        if not self.table_exists(table_name):
            raise TableNotFound("Table with name <%s> not found."
                                % table_name)
        return r.table(table_name)

    def execute(self, stmt):
        """Executes statement against the database.

        @param stmt: statement to run against db.
        @returns Result obtained from running the statement.
        @raises ValueError: The statement is not an instance of
        redirectdb.ast.RqlQuery as expected.
        """
        if not isinstance(stmt, RqlQuery):
            raise ValueError('Expecting <RqlQuery> instance, got <%s>'
                             % type(stmt))

        return stmt.run(self.connection)



from scrapy.item import Item
from scrapy.exceptions import NotConfigured

from scrapy_rethinkdb.driver import RethinkDBDriver


class RethinkDBPipeline(object):
    @classmethod
    def from_crawler(cls, crawler):
        """Gets settings for the pipeline from the crawler.
        @param crawler: crawler
        """
        settings = crawler.settings

        # get relevant settings for the pipeline
        connection_settings = settings.get('RETHINKDB_CONNECTION', {})
        table_name = settings.get('RETHINKDB_TABLE', None)
        insert_options = settings.get('RETHINKDB_INSERT_OPTIONS', {})

        # creates driver instance
        driver = RethinkDBDriver(connection_settings)

        return cls(driver, table_name, insert_options)

    def __init__(self, driver, table_name, insert_options):
        """Rethinkdb pipeline for Scrapy processed items.

        @param driver: rethinkdb driver
        @param table_name: rethinkdb driver
        @param insert_options: options to be used when inserting documents

        @raises NotConfigured: if any of the constructor arguments wasn't
        supplied.
        """

        # check if driver not provided
        if not driver:
            raise NotConfigured("Driver not provided.")
        else:
            self.driver = driver

        # check if table name provided
        if not table_name:
            raise NotConfigured("Table name not provided.")
        else:
            # get table reference from driver
            self.table = self.driver.get_table(table_name)

        # check if insert options not provided
        if insert_options is None:
            raise NotConfigured("Insert options not provided.")
        else:
            self.insert_options = insert_options

    def process_item(self, item, spider):
        """Processes item.

        @param item: the item scraped
        @param spider: the spider which scraped the item
        @returns Item to be processed by the next pipelines (if any).
        """
        if not isinstance(item, Item):
            spider.log.msg('Item not valid for RethinkDBPipeline <%s>. '
                           'Ignoring item.' % item,
                           level=spider.log.WARNING)
            return item

        self.before_insert(item)
        document = self.get_document(item)
        insert_stmt = self.table.insert(document, **self.insert_options)
        insert_result = self.driver.execute(insert_stmt)
        self.after_insert(item, insert_result)

        return item

    def get_document(self, item):
        """Gets document from the item. By default returns the attribute
        _values from the item.

        @param item: the item scraped
        @returns Document that represents the item.
        """
        return item._values

    def before_insert(self, item):
        """
        Extension point to change the item/ drop it before inserting it in
        the database.

        @param item: the item scraped
        """
        pass

    def after_insert(self, item, insert_result):
        """
        Extension point to change the item/ drop it with the result from
        inserting it in the database.

        @param item: the item scraped
        @param insert_result: result from insert
        """
        pass



"""
Pipeline to insert documents to a RethinkDB table.

The minimal configuration requires defining the pipeline in the item pipelines
and a table name.

    ITEM_PIPELINES = [
      'scrapy_rethinkdb.RethinkDBPipeline',
    ]

    RETHINKDB_TABLE = 'items'

The connection will be created with the defaults of RethinkDB's python
driver (see http://www.rethinkdb.com/api/python/connect/) unless is
overwritten as in the following example to use the database called scrapydb:

    RETHINKDB_CONNECTION = {
        'db': 'scrapydb'
    }

The insertion will be done with the default options of RethinkDB's python
driver (see http://www.rethinkdb.com/api/python/insert/) unless is
overwritten as in the following example to insert or update the item:

    RETHINKDB_INSERT_OPTIONS = {
        'upsert': False
    }

The document inserted by default is the dictionary with values from the item.
This can be changed by extending the pipeline and overriding the method
get_document.

If any additional behaviour is needed before/after the insertion, such as
changing the item, dropping it, etc, this can be done by extending the pipeline
and overriding the method(s) before_insert/after_insert.
"""

from .pipeline import RethinkDBPipeline



import unittest
from mock import Mock, patch, call

from scrapy_rethinkdb.driver import RethinkDBDriver, RqlQuery, TableNotFound


class RethinkDBDriverTest(unittest.TestCase):

    def setUp(self):
        # mocks
        self.stmt_mock = Mock(spec=RqlQuery)
        self.table_query_mock = Mock(spec=RqlQuery)

        # global patch for the official driver
        self.r_patcher = patch('scrapy_rethinkdb.driver.r')
        self.r = self.r_patcher.start()

        # default driver under test and connection call expected
        self.driver = RethinkDBDriver({'param1': 1, 'param2': 2})
        self.connection_call = call(param1=1, param2=2)

        #table name
        self.table_name = 'table'

    def tearDown(self):
        # stop global patchers
        self.r_patcher.stop()

    def test_init_invalid_settings_type(self):
        self.assertRaises(ValueError, RethinkDBDriver, None)
        self.assertRaises(ValueError, RethinkDBDriver, Mock())

    def test_connection_first_connection(self):
        connection = self.driver.connection
        self.assertEqual(self.r.connect.return_value, connection)
        self.assertEqual([self.connection_call], self.r.connect.call_args_list)

    def test_connection_second_connection(self):
        connection1 = self.driver.connection
        connection2 = self.driver.connection
        self.assertEqual(self.r.connect.return_value, connection1)
        self.assertEqual(self.r.connect.return_value, connection2)
        self.assertEqual([self.connection_call], self.r.connect.call_args_list)

    def test_connection_exception(self):
        self.r.connect.side_effect = Exception
        self.assertRaises(Exception, lambda: self.driver.connection)

    def test_execute_invalid_stmt(self):
        self.assertRaises(ValueError, self.driver.execute, Mock())

    def test_execute_exception(self):
        self.stmt_mock.run.side_effect = Exception
        self.assertRaises(Exception, self.driver.execute, self.stmt_mock)
        self.stmt_mock.run.assert_called_once_with(self.r.connect.return_value)

    def test_execute_success(self):
        result = self.driver.execute(self.stmt_mock)
        self.stmt_mock.run.assert_called_once_with(self.r.connect.return_value)
        self.assertEqual(self.stmt_mock.run.return_value, result)

    def test_table_exists_empty_database(self):
        self.r.table_list.return_value = self.stmt_mock
        self.r.table_list.return_value.run.return_value = []
        self.assertFalse(self.driver.table_exists(self.table_name))
        self.r.table_list.assert_called_once_with()
        self.stmt_mock.run.assert_called_once_with(self.r.connect.return_value)

    def test_table_exists_no_table_in_database(self):
        self.r.table_list.return_value = self.stmt_mock
        self.r.table_list.return_value.run.return_value = [
            'not_' + self.table_name,
            'really_not_' + self.table_name
        ]
        self.assertFalse(self.driver.table_exists(self.table_name))
        self.r.table_list.assert_called_once_with()
        self.stmt_mock.run.assert_called_once_with(self.r.connect.return_value)

    def test_table_exists_only_table_in_database(self):
        self.r.table_list.return_value = self.stmt_mock
        self.r.table_list.return_value.run.return_value = [self.table_name]
        self.assertTrue(self.driver.table_exists(self.table_name))
        self.r.table_list.assert_called_once_with()
        self.stmt_mock.run.assert_called_once_with(self.r.connect.return_value)

    def test_table_exists_table_in_database(self):
        self.r.table_list.return_value = self.stmt_mock
        self.r.table_list.return_value.run.return_value = [
            'not_' + self.table_name,
            'really_not_' + self.table_name,
            self.table_name
        ]
        self.assertTrue(self.driver.table_exists(self.table_name))
        self.r.table_list.assert_called_once_with()
        self.stmt_mock.run.assert_called_once_with(self.r.connect.return_value)

    def test_get_table_empty_database(self):
        self.r.table_list.return_value = self.stmt_mock
        self.r.table_list.return_value.run.return_value = []

        self.assertRaises(TableNotFound,
                          self.driver.get_table, self.table_name)

        self.r.table_list.assert_called_once_with()
        self.stmt_mock.run.assert_called_once_with(self.r.connect.return_value)
        self.assertFalse(self.r.table.called)

    def test_get_table_no_table_in_database(self):

        self.r.table_list.return_value = self.stmt_mock
        self.r.table_list.return_value.run.return_value = [
            'not_' + self.table_name,
            'really_not_' + self.table_name
        ]
        self.assertRaises(TableNotFound,
                          self.driver.get_table, self.table_name)
        self.r.table_list.assert_called_once_with()
        self.stmt_mock.run.assert_called_once_with(self.r.connect.return_value)
        self.assertFalse(self.r.table.called)

    def test_get_table_only_table_in_database(self):
        self.r.table_list.return_value = self.stmt_mock
        self.r.table_list.return_value.run.return_value = [self.table_name]
        self.r.table.return_value = self.table_query_mock

        table = self.driver.get_table(self.table_name)
        self.assertEqual(self.r.table.return_value, table)

        self.r.table_list.assert_called_once_with()
        self.r.table.assert_called_once_with(self.table_name)
        self.stmt_mock.run.assert_called_once_with(self.r.connect.return_value)

    def test_get_table_table_in_database(self):
        self.r.table_list.return_value = self.stmt_mock
        self.r.table_list.return_value.run.return_value = [
            'not_' + self.table_name,
            'really_not_' + self.table_name,
            self.table_name
        ]
        self.r.table.return_value = self.table_query_mock

        table = self.driver.get_table(self.table_name)
        self.assertEqual(self.r.table.return_value, table)

        self.r.table_list.assert_called_once_with()
        self.r.table.assert_called_once_with(self.table_name)
        self.stmt_mock.run.assert_called_once_with(self.r.connect.return_value)



import unittest
from mock import Mock, MagicMock, patch
from itertools import combinations_with_replacement

from scrapy_rethinkdb.pipeline import RethinkDBPipeline, NotConfigured, Item


class RethinkDBPipelineTest(unittest.TestCase):

    def setUp(self):

        # default pipeline under test
        self.driver = Mock()
        self.table_name = Mock()
        self.insert_options = MagicMock()
        self.pipeline = RethinkDBPipeline(
            self.driver, self.table_name, self.insert_options
        )

        # patch for driver
        self.driver_patcher = patch('scrapy_rethinkdb.pipeline.'
                                    'RethinkDBDriver')
        # patcher for the pipeline constructor
        self.pipeline_cls_patcher = patch('scrapy_rethinkdb.pipeline.'
                                          'RethinkDBPipeline.__init__')

        # returns iterator for all possible combinations for 3 arguments
        # which values can be either a Mock or None
        self.init_mocks_iter = lambda: \
            combinations_with_replacement((None, Mock()), 3)

        # returns settings dictionary
        self.get_pipeline_settings = \
            lambda conn_sett, table_name, insert_options: \
            {'RETHINKDB_TABLE': table_name,
             'RETHINKDB_CONNECTION': conn_sett,
             'RETHINKDB_INSERT_OPTIONS': insert_options}

    def test_init_not_configured(self):
        # asserts constructor will raise NotConfigured id any of the arguments
        # is None

        comb_iter = self.init_mocks_iter()
        for driver, table_name, insert_options in comb_iter:
            if not driver or not table_name or not insert_options:
                self.assertRaises(NotConfigured, RethinkDBPipeline,
                                  driver, table_name, insert_options)

    def test_init_empty_table_name(self):
        # asserts constructor will raise NotConfigured if table_name is empty
        self.assertRaises(NotConfigured, RethinkDBPipeline, Mock(), '', Mock())

    def test_init_configured(self):
        # asserts that the default pipeline under test tried to get the table

        self.assertEqual(self.pipeline.table,
                         self.driver.get_table.return_value)
        self.driver.get_table.assert_called_once_with(self.table_name)

    def test_from_crawler_configured(self):
        # asserts that from_crawler will return a pipeline instance if the
        # constructor returns None, as expected

        crawler = Mock()

        comb_iter = self.init_mocks_iter()
        for conn_sett, table_name, insert_options in comb_iter:
            crawler.settings = self.get_pipeline_settings(
                conn_sett, table_name, insert_options
            )

            with self.pipeline_cls_patcher as pipeline_cls, \
                    self.driver_patcher as driver_klass:
                pipeline_cls.return_value = None

                pipeline = RethinkDBPipeline.from_crawler(crawler)
                self.assertIsInstance(pipeline, RethinkDBPipeline)

                driver_klass.assert_called_once_with(conn_sett)
                pipeline_cls.assert_called_once_with(
                    driver_klass.return_value, table_name, insert_options
                )

    def test_from_crawler_not_configured(self):
        # asserts that from_crawler will raise NotConfigured the
        # constructor raises NotConfigured exception
        crawler = Mock()

        for conn_sett, table_name, insert_options in self.init_mocks_iter():
            crawler.settings = self.get_pipeline_settings(
                conn_sett, table_name, insert_options
            )

            with self.pipeline_cls_patcher as pipeline_cls, \
                    self.driver_patcher as driver_klass:
                pipeline_cls.side_effect = NotConfigured

                self.assertRaises(NotConfigured,
                                  RethinkDBPipeline.from_crawler,
                                  crawler)

                driver_klass.assert_called_once_with(conn_sett)
                pipeline_cls.assert_called_once_with(
                    driver_klass.return_value, table_name, insert_options
                )

    def test_process_item_not_an_item(self):
        # asserts that a non-item is just returned

        spider = Mock()
        # mocking before_insert to check that it won't be called in this test
        self.pipeline.before_insert = Mock()
        self.pipeline.process_item(Mock(), spider)
        self.assertTrue(spider.log.msg.called)
        self.assertFalse(self.pipeline.before_insert.called)

    def test_process_item_success(self):
        # asserts that a item is processed

        item = Mock(spec=Item)
        item._values = {}

        # mocking extension points
        self.pipeline.before_insert = Mock()
        self.pipeline.after_insert = Mock()

        self.pipeline.process_item(item, Mock())

        self.pipeline.before_insert.assert_called_once_with(
            item
        )
        self.pipeline.table.insert.assert_called_once_with(
            item._values
        )
        self.pipeline.driver.execute.assert_called_once_with(
            self.pipeline.table.insert.return_value
        )
        self.pipeline.after_insert.assert_called_once_with(
            item, self.pipeline.driver.execute.return_value
        )



from setuptools import setup, find_packages
from pip.req import parse_requirements

setup(
    name='scrapy-rethinkdb',
    version='0.0.4',
    author='sprij',
    author_email='s.rijo@yahoo.com',
    description='Scrapy pipeline for rethinkdb.',
    keywords="scrapy pipeline rethinkdb",
    url='https://github.com/sprij/scrapy-rethinkdb',
    packages=find_packages(),
    include_package_data=True,
    long_description=__doc__,
    install_requires=[
        str(install_req.req)
        for install_req in parse_requirements('requirements.txt')
    ],
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Topic :: Utilities",
        'Programming Language :: Python',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: GNU General Public License v2 (GPLv2)',
    ]
)
